#ifndef WIN32
#include "config.h"
#endif

#include "common.h"
#include "application.h"
#include "screen.h"
#include "utils.h"
#include "debug.h"
#include "screen_layer.h"
#include "monitor.h"
#include "controller.h"
#include "resource.h"
#ifdef WIN32
#include "spicexipc.h"
#include "gdicanvas.h"
#endif
#include "platform.h"
#include "cairo_canvas.h"
#include "gl_canvas.h"
#include "quic.h"
#include "mutex.h"
#include "rect.h"

#define STICKY_KEY_PIXMAP ALT_IMAGE_RES_ID
#define STICKY_KEY_TIMEOUT 750


#define CA_FILE_NAME "spice_truststore.pem"

#ifdef CAIRO_CANVAS_CACH_IS_SHARED
mutex_t cairo_surface_user_data_mutex;
#endif

#ifdef WIN32
static const WCHAR default_menu_resource[] =
    L"0\r0\r-\r1\n"
    L"0\r4864\rSend Ctrl+Alt+Del\tCtrl+Alt+End\r0\n"
    L"0\r5120\rToggle full screen\tShift+F11\r0\n"
    L"0\r2\rSpecial keys\r4\n"
        L"2\r5376\rSend Shift+F11\r0\n"
        L"2\r5632\rSend Shift+F12\r0\n"
        L"2\r5888\rSend Ctrl+Alt+End\r0\n"
    ;
#endif


SyncEvent::SyncEvent()
    : _err (false)
    , _ready (false)
{
}

SyncEvent::~SyncEvent()
{
}

void SyncEvent::responce(Application& application)
{
    try {
        do_responce(application);
    } catch  (Exception& e) {
         WARN(fmt("unhandle exception: %s") % e.what());
        _err = true;
    } catch ( ... ) {
        _err = true;
    }
    Lock lock(_mutex);
    _ready = true;
    _condition.notify_one();
}

void SyncEvent::wait()
{
    //todo: process event if on main thread
    Lock lock(_mutex);
    while (!_ready) {
        _condition.wait(lock);
    }
}

void ConnectedEvent::responce(Application& application)
{
    application.on_connected();
}

void DisconnectedEvent::responce(Application& application)
{
#ifdef RED_DEBUG
    application.show_splash(0);
#else
    application.quit(SPICEC_ERROR_CODE_SUCCESS);
#endif
}

void CoonnectionError::responce(Application& application)
{
#ifdef RED_DEBUG
    application.show_splash(0);
#else
    application.quit(_error_code);
#endif
}

void ErrorEvent::responce(Application& application)
{
    application.quit(SPICEC_ERROR_CODE_ERROR);
}

void MonitorsQuery::do_responce(Application& application)
{
    Monitor* mon;
    int i = 0;

    while ((mon = application.find_monitor(i++))) {
        MonitorInfo info;
        info.size = mon->get_size();
        info.depth = mon->get_depth();
        info.position = mon->get_position();
        _monitors.push_back(info);
    }
}

class GUILayer: public ScreenLayer{
public:
    GUILayer();

    virtual void copy_pixels(const QRegion& dest_region, RedDrawable& dest_dc);

    void set_splash_mode();
    void set_info_mode();
    void set_sticky(bool is_on);
    virtual void on_size_changed();

private:
    void draw_splash(const QRegion& dest_region, RedDrawable& dest);
    void draw_info(const QRegion& dest_region, RedDrawable& dest);

private:
    ImageFromRes _splash_pixmap;
    AlphaImageFromRes _info_pixmap;
    AlphaImageFromRes _sticky_pixmap;
    Point _splash_pos;
    Point _info_pos;
    Point _sticky_pos;
    Rect _sticky_rect;
    bool _splash_mode;
    bool _sticky_on;
    RecurciveMutex _update_lock;
};

GUILayer::GUILayer()
    : ScreenLayer(SCREEN_LAYER_GUI, false)
    , _splash_pixmap (SPLASH_IMAGE_RES_ID)
    , _info_pixmap (INFO_IMAGE_RES_ID)
    , _sticky_pixmap (STICKY_KEY_PIXMAP)
    , _splash_mode (false)
    , _sticky_on (false)
{

}

void GUILayer::draw_splash(const QRegion& dest_region, RedDrawable& dest)
{
    ASSERT(!_sticky_on);
    for (int i = 0; i < (int)dest_region.num_rects; i++) {
        Rect* r = &dest_region.rects[i];
        dest.copy_pixels(_splash_pixmap, r->left - _splash_pos.x, r->top - _splash_pos.y, *r);
    }
}

void GUILayer::draw_info(const QRegion& dest_region, RedDrawable& dest)
{
    for (int i = 0; i < (int)dest_region.num_rects; i++) {
        Rect* r = &dest_region.rects[i];
        /* is rect inside sticky region or info region? */
        if (_sticky_on && rect_intersects(*r, _sticky_rect)) {
            dest.blend_pixels(_sticky_pixmap, r->left - _sticky_pos.x, r->top - _sticky_pos.y, *r);
        } else {
            dest.blend_pixels(_info_pixmap, r->left - _info_pos.x, r->top - _info_pos.y, *r);
        }
    }
}

void GUILayer::copy_pixels(const QRegion& dest_region, RedDrawable& dest_dc)
{
    RecurciveLock lock(_update_lock);
    if (_splash_mode) {
        draw_splash(dest_region, dest_dc);
    } else {
        draw_info(dest_region, dest_dc);
    }
}

void GUILayer::set_splash_mode()
{
    RecurciveLock lock(_update_lock);
    Point size = _splash_pixmap.get_size();
    Point screen_size = screen()->get_size();
    Rect r;

    _splash_pos.y = r.top = (screen_size.y - size.y) / 2;
    _splash_pos.x = r.left = (screen_size.x - size.x) / 2;
    r.bottom = r.top + size.y;
    r.right = r.left + size.x;
    _splash_mode = true;
    lock.unlock();
    set_rect_area(r);
    ASSERT(!_sticky_on);
}

void GUILayer::set_info_mode()
{
    RecurciveLock lock(_update_lock);
    Point size = _info_pixmap.get_size();
    Point screen_size = screen()->get_size();
    Rect r;

    r.left = (screen_size.x - size.x) / 2;
    r.right = r.left + size.x;
    _info_pos.x = r.right - size.x;
    _info_pos.y = r.top = 0;
    r.bottom = r.top + size.y;
    _splash_mode = false;
    lock.unlock();
    set_rect_area(r);

    set_sticky(_sticky_on);
}

void GUILayer::set_sticky(bool is_on)
{
    RecurciveLock lock(_update_lock);
    if (!_sticky_on && !is_on) {
        return;
    }

    Point size = _sticky_pixmap.get_size();
    Point screen_size = screen()->get_size();

    _sticky_on = is_on;
    if (_sticky_on) {
        _sticky_pos.x = (screen_size.x - size.x) / 2;
        _sticky_pos.y = screen_size.y * 2 / 3;
        _sticky_rect.left = _sticky_pos.x;
        _sticky_rect.top = _sticky_pos.y;
        _sticky_rect.right = _sticky_rect.left + size.x;
        _sticky_rect.bottom = _sticky_rect.top + size.y;
        add_rect_area(_sticky_rect);
        invalidate();
    } else {
        remove_rect_area(_sticky_rect);
    }
}

void GUILayer::on_size_changed()
{
    set_info_mode();
}

static InputsHandler default_inputs_handler;

void Application::init()
{
    srand((unsigned)time(NULL));
    cairo_canvas_init();
#ifdef USE_OGL
    gl_canvas_init();
#endif
    quic_init();
#ifdef WIN32
    gdi_canvas_init();
#endif

#ifdef CAIRO_CANVAS_CACH_IS_SHARED
    MUTEX_INIT(cairo_surface_user_data_mutex);
#endif

    Platform::init();
    RedWindow::init();
}

void Application::cleanup()
{
    RedWindow::cleanup();
}

Application::Application()
    : _client (*this)
    , _main_screen (NULL)
    , _quitting (false)
    , _active (false)
    , _full_screen (false)
    , _changing_screens (false)
    , _send_cad (false)
    , _exit_code (0)
    , _events_gen (0)
    , _active_screen (NULL)
    , _gui_layer(new GUILayer())
    , _inputs_handler (&default_inputs_handler)
    , _monitors (NULL)
    , _foreign_menu (this)
    , _splash_mode (true)
    , _sys_key_intercept_mode (false)
    , _controller_is_up (false)
#ifdef WIN32
    , _spice_ipc (NULL)
    , _menu_resource (default_menu_resource)
    , _no_taskmgr (false)
    , _canvas_type ("gdi")
#else
    , _canvas_type ("cairo")
#endif
{
    LOG_TRACE("");

    init_monitors();
    init_key_table();
    _main_screen = get_screen(0);
    _main_screen->attach_layer(*_gui_layer);
    _gui_layer->set_splash_mode();
    _client.register_channel_factory(DisplayChannel::Factory());
    _client.register_channel_factory(CursorChannel::Factory());
    _client.register_channel_factory(InputsChannel::Factory());

    _actions[SpiceConfiguration::ActionToggleFullScreen] = &Application::toggle_full_screen;
    _actions[SpiceConfiguration::ActionReleaseCursor] = &Application::release_capture;
#ifdef RED_DEBUG
    _actions[SpiceConfiguration::ActionConnect] = &Application::do_connect;
    _actions[SpiceConfiguration::ActionDisconnect] = &Application::do_disconnect;
#endif // RED_DEBUG
    Platform::set_event_listener(this);
    Platform::set_display_mode_listner(this);

    _sticky_info.trace_is_on = false;
    _sticky_info.sticky_mode = false;
    _sticky_info.key_first_down = false;
    _sticky_info.key_down = false;
    _sticky_info.key  = REDKEY_INVALID;
    _sticky_info.timer = Platform::create_interval_timer(sticky_timeout_proc, this);

    if (_sticky_info.timer == INVALID_TIMER) {
        throw Exception(fmt("%s: invalid sticky timer") % __FUNCTION__ );
    }
}

Application::~Application()
{
    Platform::destroy_interval_timer(_sticky_info.timer);
    _main_screen->detach_layer(*_gui_layer);
    _main_screen->unref();
    destroy_monitors();
}

void Application::set_inputs_handler(InputsHandler& handler)
{
    unpress_all();
    _inputs_handler = &handler;
    if (_active) {
        handler.on_focus_in();
    }
}

void Application::remove_inputs_handler(InputsHandler& handler)
{
    if (_inputs_handler != &handler) {
        return;
    }
    _inputs_handler = &default_inputs_handler;
}

void Application::process_events()
{
    _events_gen++;
    for (;;) {
        Event* event;
        Lock lock(_events_lock);
        if (_events.empty()) {
            return;
        }
        event = _events.front();
        if (event->_generation == _events_gen) {
            Platform::wakeup();
            return;
        }
        _events.pop_front();
        lock.unlock();
        event->responce(*this);
        event->unref();
    }
}

void Application::push_event(Event* event)
{
    Lock lock(_events_lock);
    bool notify = _events.empty();
    _events.push_back(event);
    event->_generation = _events_gen;
    event->ref();
    if (notify) {
        Platform::wakeup();
    }
}

int Application::message_loop()
{
    for (;;) {
        Platform::wait_events();
        if (Platform::process_events()) {
            _quitting = true;
            break;
        }
        process_events();
    }
    return _exit_code;
}

void Application::abort()
{
    Platform::set_event_listener(NULL);
    Platform::set_display_mode_listner(NULL);
    unpress_all();
    while (!_client.abort()) {
        process_events();
        Platform::yield();
    }
}

class AutoAbort {
public:
    AutoAbort(Application& app) : _app(app) {}
    ~AutoAbort() { _app.abort();}
private:
    Application& _app;
};

void Application::set_ssl_config(SpiceConfiguration& spice_config)
{
    bool ssl_encrypt = false;
    const bool *encrypted_channels = spice_config._encrypt_channels;
    std::string default_CA_file;
    Platform::get_spice_config_dir(default_CA_file);
    default_CA_file += CA_FILE_NAME;

    for (int i=0; i < RED_CHANNEL_END; i++) {
        if (encrypted_channels[i]) {
            peer_connect_options& peer_config = _peers_configuration[i];
            ssl_encrypt = true;
            peer_config.SetSSLParameters(true, "cert.pem", "key.pem", "123456",
                                         spice_config._cipher.c_str(),
                                         spice_config._sport);

            if (!_controller_is_up) {
                int auth_flags = SERVER_AUTH_NAME ;
                std::string ca_file;

                if (!spice_config._host_cert_subject.empty()) {
                    auth_flags |= SERVER_AUTH_SUBJECT;
                }

                ca_file = spice_config._CA_file.empty() ? default_CA_file : spice_config._CA_file;
                peer_config.SetSSLServerAuthParameters(auth_flags,
                                                       NULL, 0,
                                                       ca_file.c_str(),
                                                       spice_config._host_cert_subject);
            } else {
                if (!spice_config._CA_file.empty()) {
                    int auth_flags = spice_config._host_cert_subject.empty() ? SERVER_AUTH_NAME :
                                                                               SERVER_AUTH_SUBJECT;
                    peer_config.SetSSLServerAuthParameters(auth_flags,
                                                           NULL, 0,
                                                           spice_config._CA_file.c_str(),
                                                           spice_config._host_cert_subject);
                }
            }
        } else {
            _peers_configuration[i].SetSSLParameters(false, NULL, NULL, NULL, NULL,
                                                     spice_config._port);
        }
    }

    if (ssl_encrypt) {
        SSL_library_init();
        SSL_load_error_strings();
    }
}

void Application::connect()
{
    _client.connect(_peers_configuration);
}

int Application::run()
{
    try {
        AutoAbort auto_abort(*this);

#ifdef WIN32
        _client.connect(_peers_configuration);
#else
        if (!_external_controller.get()) {
            _client.connect(_peers_configuration);
        }
#endif
        _exit_code = message_loop();
    } catch (...) {
        throw;
    }
    return _exit_code;
}

void Application::quit(int exit_code)
{
    if (!_quitting) {
        _quitting = true;
        _exit_code = exit_code;
        Platform::send_quit_request();
    }
}

RedScreen* Application::find_screen(int id)
{
    if ((int)_screens.size() < id + 1) {
        return NULL;
    }
    return _screens[id];
}

bool Application::release_capture()
{
    unpress_all();
    if (!_active_screen || !_active_screen->is_captured()) {
        return false;
    }
    _active_screen->relase_inputs();
    return true;
}

bool Application::do_connect()
{
    _client.connect(_peers_configuration);
    return true;
}

bool Application::do_disconnect()
{
    on_disconnecting();
    _client.disconnect();
    return true;
}

#define SCREEN_INIT_WIDTH 800
#define SCREEN_INIT_HEIGHT 600

RedScreen* Application::get_screen(int id)
{
    RedScreen *screen;

    if ((int)_screens.size() < id + 1) {
        _screens.resize(id + 1);
    }

    if (!(screen = _screens[id])) {
        Monitor* mon;

        if (_client.is_auto_display_res() && (mon = find_monitor(id))) {
            Point size = mon->get_size();
            screen = _screens[id] = new RedScreen(*this, id, _title, size.x, size.y);
        } else {
            screen = _screens[id] = new RedScreen(*this, id, _title, SCREEN_INIT_WIDTH, SCREEN_INIT_HEIGHT);
        }
#ifdef WIN32
        screen->create_menu(_menu_resource.c_str());
        screen->set_no_taskmgr(_no_taskmgr);
#endif
        if (_full_screen) {
            bool capture;

            mon = get_monitor(id);
            capture = release_capture();
            screen->set_monitor(mon);
            position_screens();
            screen->show_full_screen();
            prepare_monitors();

            if (capture) {
                _main_screen->activate();
                _main_screen->capture_inputs();
            }
        } else if (id != 0) {
            screen->show(false, _main_screen);
        }
    } else {
        screen = screen->ref();
    }

    return screen;
}

void Application::on_screen_destroyed(int id, bool was_captured)
{
    bool reposition = false;

    if ((int)_screens.size() < id + 1 || !_screens[id]) {
        throw Exception(fmt("%s: no screen") % __FUNCTION__);
    }
    if (_active_screen == _screens[id]) {
        _active_screen = NULL;
    }

    if (_full_screen && _screens[id]->has_monitor()) {
        _screens[id]->get_monitor()->restore();
        reposition = true;
    }
    _screens[id] = NULL;
    if (reposition) {
        bool capture = was_captured || release_capture();
        prepare_monitors();
        position_screens();
        if (capture) {
            _main_screen->activate();
            _main_screen->capture_inputs();
        }
    }
}

void Application::on_mouse_motion(int dx, int dy, int buttons_state)
{
    _inputs_handler->on_mouse_motion(dx, dy, buttons_state);
}

void Application::on_mouse_position(int x, int y, int buttons_state, int display_id)
{
    _inputs_handler->on_mouse_position(x, y, buttons_state, display_id);
}

void Application::on_mouse_down(int button, int buttons_state)
{
    _inputs_handler->on_mouse_down(button,  buttons_state);
}

void Application::on_mouse_up(int button, int buttons_state)
{
    _inputs_handler->on_mouse_up(button,  buttons_state);
}

void Application::init_scan_code(int index)
{
    ASSERT((index & 0x80) == 0);
    _key_table[index]._make = index;
    _key_table[index]._break = index | 0x80;
}

void Application::init_korean_scan_code(int index)
{
    _key_table[index]._make = index;
    _key_table[index]._break = index;
}

void Application::init_escape_scan_code(int index)
{
    ASSERT(((index - REDKEY_ESCAPE_BASE) & 0x80) == 0);
    _key_table[index]._make = 0xe0 | ((index - REDKEY_ESCAPE_BASE) << 8);
    _key_table[index]._break = _key_table[index]._make | 0x8000;
}

void Application::init_pause_scan_code()
{
    _key_table[REDKEY_PAUSE]._make = 0x451de1;
    _key_table[REDKEY_PAUSE]._break = 0xc59de1;
}

void Application::init_key_table()
{
    memset(_key_table, 0, sizeof(_key_table));
    _num_keys_pressed = 0;
    init_scan_code(REDKEY_ESCAPE);
    init_scan_code(REDKEY_1);
    init_scan_code(REDKEY_2);
    init_scan_code(REDKEY_3);
    init_scan_code(REDKEY_4);
    init_scan_code(REDKEY_5);
    init_scan_code(REDKEY_6);
    init_scan_code(REDKEY_7);
    init_scan_code(REDKEY_8);
    init_scan_code(REDKEY_9);
    init_scan_code(REDKEY_0);
    init_scan_code(REDKEY_MINUS);
    init_scan_code(REDKEY_EQUALS);
    init_scan_code(REDKEY_BACKSPACE);
    init_scan_code(REDKEY_TAB);
    init_scan_code(REDKEY_Q);
    init_scan_code(REDKEY_W);
    init_scan_code(REDKEY_E);
    init_scan_code(REDKEY_R);
    init_scan_code(REDKEY_T);
    init_scan_code(REDKEY_Y);
    init_scan_code(REDKEY_U);
    init_scan_code(REDKEY_I);
    init_scan_code(REDKEY_O);
    init_scan_code(REDKEY_P);
    init_scan_code(REDKEY_L_BRACKET);
    init_scan_code(REDKEY_R_BRACKET);
    init_scan_code(REDKEY_ENTER);
    init_scan_code(REDKEY_L_CTRL);
    init_scan_code(REDKEY_A);
    init_scan_code(REDKEY_S);
    init_scan_code(REDKEY_D);
    init_scan_code(REDKEY_F);
    init_scan_code(REDKEY_G);
    init_scan_code(REDKEY_H);
    init_scan_code(REDKEY_J);
    init_scan_code(REDKEY_K);
    init_scan_code(REDKEY_L);
    init_scan_code(REDKEY_SEMICOLON);
    init_scan_code(REDKEY_QUOTE);
    init_scan_code(REDKEY_BACK_QUOTE);
    init_scan_code(REDKEY_L_SHIFT);
    init_scan_code(REDKEY_BACK_SLASH);
    init_scan_code(REDKEY_Z);
    init_scan_code(REDKEY_X);
    init_scan_code(REDKEY_C);
    init_scan_code(REDKEY_V);
    init_scan_code(REDKEY_B);
    init_scan_code(REDKEY_N);
    init_scan_code(REDKEY_M);
    init_scan_code(REDKEY_COMMA);
    init_scan_code(REDKEY_PERIOD);
    init_scan_code(REDKEY_SLASH);
    init_scan_code(REDKEY_R_SHIFT);
    init_scan_code(REDKEY_PAD_MULTIPLY);
    init_scan_code(REDKEY_L_ALT);
    init_scan_code(REDKEY_SPACE);
    init_scan_code(REDKEY_CAPS_LOCK);
    init_scan_code(REDKEY_F1);
    init_scan_code(REDKEY_F2);
    init_scan_code(REDKEY_F3);
    init_scan_code(REDKEY_F4);
    init_scan_code(REDKEY_F5);
    init_scan_code(REDKEY_F6);
    init_scan_code(REDKEY_F7);
    init_scan_code(REDKEY_F8);
    init_scan_code(REDKEY_F9);
    init_scan_code(REDKEY_F10);
    init_scan_code(REDKEY_NUM_LOCK);
    init_scan_code(REDKEY_SCROLL_LOCK);
    init_scan_code(REDKEY_PAD_7);
    init_scan_code(REDKEY_PAD_8);
    init_scan_code(REDKEY_PAD_9);
    init_scan_code(REDKEY_PAD_MINUS);
    init_scan_code(REDKEY_PAD_4);
    init_scan_code(REDKEY_PAD_5);
    init_scan_code(REDKEY_PAD_6);
    init_scan_code(REDKEY_PAD_PLUS);
    init_scan_code(REDKEY_PAD_1);
    init_scan_code(REDKEY_PAD_2);
    init_scan_code(REDKEY_PAD_3);
    init_scan_code(REDKEY_PAD_0);
    init_scan_code(REDKEY_PAD_POINT);

    init_scan_code(REDKEY_EUROPEAN);
    init_scan_code(REDKEY_F11);
    init_scan_code(REDKEY_F12);

    init_scan_code(REDKEY_JAPANESE_HIRAGANA_KATAKANA);
    init_scan_code(REDKEY_JAPANESE_BACKSLASH);
    init_scan_code(REDKEY_JAPANESE_HENKAN);
    init_scan_code(REDKEY_JAPANESE_MUHENKAN);
    init_scan_code(REDKEY_JAPANESE_YEN);

    init_korean_scan_code(REDKEY_KOREAN_HANGUL);
    init_korean_scan_code(REDKEY_KOREAN_HANGUL_HANJA);

    init_escape_scan_code(REDKEY_ESCAPE_BASE);
    init_escape_scan_code(REDKEY_PAD_ENTER);
    init_escape_scan_code(REDKEY_R_CTRL);
    init_escape_scan_code(REDKEY_FAKE_L_SHIFT);
    init_escape_scan_code(REDKEY_PAD_DIVIDE);
    init_escape_scan_code(REDKEY_FAKE_R_SHIFT);
    init_escape_scan_code(REDKEY_CTRL_PRINT_SCREEN);
    init_escape_scan_code(REDKEY_R_ALT);
    init_escape_scan_code(REDKEY_CTRL_BREAK);
    init_escape_scan_code(REDKEY_HOME);
    init_escape_scan_code(REDKEY_UP);
    init_escape_scan_code(REDKEY_PAGEUP);
    init_escape_scan_code(REDKEY_LEFT);
    init_escape_scan_code(REDKEY_RIGHT);
    init_escape_scan_code(REDKEY_END);
    init_escape_scan_code(REDKEY_DOWN);
    init_escape_scan_code(REDKEY_PAGEDOWN);
    init_escape_scan_code(REDKEY_INSERT);
    init_escape_scan_code(REDKEY_DELETE);
    init_escape_scan_code(REDKEY_LEFT_CMD);
    init_escape_scan_code(REDKEY_RIGHT_CMD);
    init_escape_scan_code(REDKEY_MENU);

    init_pause_scan_code();
}

inline uint32_t Application::get_make_scan_code(RedKey key)
{
    return _key_table[key]._make;
}

inline uint32_t Application::get_break_scan_code(RedKey key)
{
    return _key_table[key]._break;
}

void Application::unpress_all()
{
    reset_sticky();
    for ( int i = 0; i < REDKEY_NUM_KEYS; i++) {
        if (_key_table[i].press) {
            uint32_t scan_code = get_break_scan_code((RedKey)i);
            ASSERT(scan_code);
            _inputs_handler->on_key_up(scan_code);
            unpress_key((RedKey)i);
        }
    }
}

void Application::on_connected()
{
}

void Application::on_disconnecting()
{
    release_capture();
}

#ifdef WIN32
void Application::on_sys_command(WPARAM wparam, LPARAM lparam)
{
    WPARAM wParam = wparam & 0xFFF0;

    // don't log windows events like SC_MOVE and SC_MOUSEMENU.
    if ((wParam & 0xF000) != 0xF000) {
        LOG_DEBUG("WM_SYSCOMMAND - 0x" << std::hex << wParam);
    }

    switch (wParam)
    {
    case ID_SYS_MENU_CONNECT:
        _client.connect(_peers_configuration);
        break;
    case ID_SYS_MENU_DISCONNECT:
        _client.disconnect();
        break;
    case ID_SYS_MENU_SEND_ALT_CTRL_DEL:
        send_alt_ctl_del();
        break;
    case ID_SYS_MENU_TOGGLE_FULL_SCREEN:
        toggle_full_screen();
        break;
    case ID_SYS_MENU_SEND_TOGGLE_KEYS:
        send_action_hotkeys(SpiceConfiguration::ActionToggleFullScreen);
        break;
    case ID_SYS_MENU_SEND_RELEASE_KEYS:
        send_action_hotkeys(SpiceConfiguration::ActionReleaseCursor);
        break;
    case ID_SYS_MENU_SEND_CTRL_ALT_END:
        send_ctrl_alt_end();
        break;
    default:
        if (wParam >= ID_SYS_MENU_FOREIGN_MENU_BEGIN && wParam <= ID_SYS_MENU_FOREIGN_MENU_END) {
            _active_screen->on_foreign_menu_command(wParam);
        } else if (wParam >= ID_SYS_MENU_BASE_DYNAMIC_MENU_ID && wParam <= 2*ID_SYS_MENU_BASE_DYNAMIC_MENU_ID) {
            menu_item_callback((wParam - ID_SYS_MENU_BASE_DYNAMIC_MENU_ID) >> ID_SYS_MENU_BASE_DYNAMIC_MENU_SHIFT);
        }
        break;
    }
}

#endif

void Application::on_foreign_menu_message(FrgMenuMsg *hdr, int32_t opaque_conn_ref)
{
#ifdef WIN32
    for (size_t i = 0; i < _screens.size(); ++i) {
        if (_screens[i]) {
            _screens[i]->on_foreign_menu_message(hdr, opaque_conn_ref);
        }
    }
#endif
}

void Application::on_foreign_menu_command(int32_t opaque_conn_ref, int32_t id)
{
#ifdef WIN32
    _foreign_menu.on_foreign_menu_command(opaque_conn_ref, id);
#endif
}

void Application::delete_menu(int32_t opaque_conn_ref)
{
#ifdef WIN32
    for (size_t i = 0; i < _screens.size(); ++i) {
        if (_screens[i]) {
            _screens[i]->delete_menu(opaque_conn_ref);
        }
    }
#endif
}

#ifdef REDKEY_DEBUG

static void show_red_key(RedKey key)
{

#define KEYCASE(red_key)    \
    case red_key:           \
        DBG(0, #red_key);   \
        return;

    switch (key) {
        KEYCASE(REDKEY_INVALID);
        KEYCASE(REDKEY_ESCAPE);
        KEYCASE(REDKEY_1);
        KEYCASE(REDKEY_2);
        KEYCASE(REDKEY_3);
        KEYCASE(REDKEY_4);
        KEYCASE(REDKEY_5);
        KEYCASE(REDKEY_6);
        KEYCASE(REDKEY_7);
        KEYCASE(REDKEY_8);
        KEYCASE(REDKEY_9);
        KEYCASE(REDKEY_0);
        KEYCASE(REDKEY_MINUS);
        KEYCASE(REDKEY_EQUALS);
        KEYCASE(REDKEY_BACKSPACE);
        KEYCASE(REDKEY_TAB);
        KEYCASE(REDKEY_Q);
        KEYCASE(REDKEY_W);
        KEYCASE(REDKEY_E);
        KEYCASE(REDKEY_R);
        KEYCASE(REDKEY_T);
        KEYCASE(REDKEY_Y);
        KEYCASE(REDKEY_U);
        KEYCASE(REDKEY_I);
        KEYCASE(REDKEY_O);
        KEYCASE(REDKEY_P);
        KEYCASE(REDKEY_ENTER);
        KEYCASE(REDKEY_L_BRACKET);
        KEYCASE(REDKEY_R_BRACKET);
        KEYCASE(REDKEY_L_CTRL);
        KEYCASE(REDKEY_A);
        KEYCASE(REDKEY_S);
        KEYCASE(REDKEY_D);
        KEYCASE(REDKEY_F);
        KEYCASE(REDKEY_G);
        KEYCASE(REDKEY_H);
        KEYCASE(REDKEY_J);
        KEYCASE(REDKEY_K);
        KEYCASE(REDKEY_L);
        KEYCASE(REDKEY_SEMICOLON);
        KEYCASE(REDKEY_QUOTE);
        KEYCASE(REDKEY_BACK_QUOTE);
        KEYCASE(REDKEY_L_SHIFT);
        KEYCASE(REDKEY_BACK_SLASH);
        KEYCASE(REDKEY_Z);
        KEYCASE(REDKEY_X);
        KEYCASE(REDKEY_C);
        KEYCASE(REDKEY_V);
        KEYCASE(REDKEY_B);
        KEYCASE(REDKEY_N);
        KEYCASE(REDKEY_M);
        KEYCASE(REDKEY_COMMA);
        KEYCASE(REDKEY_PERIOD);
        KEYCASE(REDKEY_SLASH);
        KEYCASE(REDKEY_R_SHIFT);
        KEYCASE(REDKEY_PAD_MULTIPLY);
        KEYCASE(REDKEY_L_ALT);
        KEYCASE(REDKEY_SPACE);
        KEYCASE(REDKEY_CAPS_LOCK);
        KEYCASE(REDKEY_F1);
        KEYCASE(REDKEY_F2);
        KEYCASE(REDKEY_F3);
        KEYCASE(REDKEY_F4);
        KEYCASE(REDKEY_F5);
        KEYCASE(REDKEY_F6);
        KEYCASE(REDKEY_F7);
        KEYCASE(REDKEY_F8);
        KEYCASE(REDKEY_F9);
        KEYCASE(REDKEY_F10);
        KEYCASE(REDKEY_NUM_LOCK);
        KEYCASE(REDKEY_SCROLL_LOCK);
        KEYCASE(REDKEY_PAD_7);
        KEYCASE(REDKEY_PAD_8);
        KEYCASE(REDKEY_PAD_9);
        KEYCASE(REDKEY_PAD_MINUS);
        KEYCASE(REDKEY_PAD_4);
        KEYCASE(REDKEY_PAD_5);
        KEYCASE(REDKEY_PAD_6);
        KEYCASE(REDKEY_PAD_PLUS);
        KEYCASE(REDKEY_PAD_1);
        KEYCASE(REDKEY_PAD_2);
        KEYCASE(REDKEY_PAD_3);
        KEYCASE(REDKEY_PAD_0);
        KEYCASE(REDKEY_PAD_POINT);
        KEYCASE(REDKEY_F11);
        KEYCASE(REDKEY_F12);
        KEYCASE(REDKEY_PAD_ENTER);
        KEYCASE(REDKEY_R_CTRL);
        KEYCASE(REDKEY_FAKE_L_SHIFT);
        KEYCASE(REDKEY_PAD_DIVIDE);
        KEYCASE(REDKEY_FAKE_R_SHIFT);
        KEYCASE(REDKEY_CTRL_PRINT_SCREEN);
        KEYCASE(REDKEY_R_ALT);
        KEYCASE(REDKEY_CTRL_BREAK);
        KEYCASE(REDKEY_HOME);
        KEYCASE(REDKEY_UP);
        KEYCASE(REDKEY_PAGEUP);
        KEYCASE(REDKEY_LEFT);
        KEYCASE(REDKEY_RIGHT);
        KEYCASE(REDKEY_END);
        KEYCASE(REDKEY_DOWN);
        KEYCASE(REDKEY_PAGEDOWN);
        KEYCASE(REDKEY_INSERT);
        KEYCASE(REDKEY_DELETE);
        KEYCASE(REDKEY_LEFT_CMD);
        KEYCASE(REDKEY_RIGHT_CMD);
        KEYCASE(REDKEY_MENU);
    default:
        DBG(0, "???");
    };
}

#endif


bool Application::press_key(RedKey key)
{
    if (_key_table[key].press) {
        return true;
    } else {
        _key_table[key].press = true;
        _num_keys_pressed++;
        return false;
    }
}

bool Application::unpress_key(RedKey key)
{
    ASSERT(!_sticky_info.key_down || !is_sticky_trace_key(key));

    if (_key_table[key].press) {
        _key_table[key].press = false;
        _num_keys_pressed--;
        ASSERT(_num_keys_pressed >= 0);
        return true;
    } else {
        return false;
    }
}

inline bool Application::is_sticky_trace_key(RedKey key)
{
    return ((key == REDKEY_L_ALT) || (key == REDKEY_R_ALT));
}

void Application::sticky_timeout_proc(void* opaque, TimerID timer)
{
    Application* app = (Application*)opaque;
    StickyInfo* sticky_info = &app->_sticky_info;
    ASSERT(app->is_sticky_trace_key(sticky_info->key));
    ASSERT(app->_key_table[sticky_info->key ].press);
    ASSERT(sticky_info->key_first_down);
    ASSERT(sticky_info->key_down);
    sticky_info->sticky_mode = true;
    DBG(0, "ON sticky");
    app->_gui_layer->set_sticky(true);
    Platform::deactivate_interval_timer(timer);
}

void Application::reset_sticky()
{
    _sticky_info.trace_is_on = !_splash_mode && _sys_key_intercept_mode;
    _sticky_info.key_first_down = false;
    Platform::deactivate_interval_timer(_sticky_info.timer);
    if (_sticky_info.sticky_mode) {
        ASSERT(_key_table[_sticky_info.key].press);
        // if it is physically down, we shouldn't unpress it
        if (!_sticky_info.key_down) {
            do_on_key_up(_sticky_info.key);
        }
        _sticky_info.sticky_mode = false;
        DBG(0, "OFF sticky");
        _gui_layer->set_sticky(false);
    }
    _sticky_info.key_down = false;

    _sticky_info.key = REDKEY_INVALID;

}

struct ModifierKey {
    int modifier;
    RedKey key;
};

ModifierKey modifier_keys[] = {
    {Platform::L_SHIFT_MODIFIER, REDKEY_L_SHIFT},
    {Platform::R_SHIFT_MODIFIER, REDKEY_R_SHIFT},
    {Platform::L_CTRL_MODIFIER, REDKEY_L_CTRL},
    {Platform::R_CTRL_MODIFIER, REDKEY_R_CTRL},
    {Platform::L_ALT_MODIFIER, REDKEY_L_ALT},
    {Platform::R_ALT_MODIFIER, REDKEY_R_ALT},
};

void Application::sync_keyboard_modifiers()
{
    uint32_t modifiers = Platform::get_keyboard_modifiers();
    for (int i = 0; i < sizeof(modifier_keys) / sizeof(modifier_keys[0]); i++) {
        if (modifiers & modifier_keys[i].modifier) {
            on_key_down(modifier_keys[i].key);
        } else {
            on_key_up(modifier_keys[i].key);
        }
    }
}

void Application::on_key_down(RedKey key)
{
    if (key <= 0 || key >= REDKEY_NUM_KEYS) {
        return;
    }

    uint32_t scan_code = get_make_scan_code(key);
    if (!scan_code) {
        WARN(fmt("no make code for %d") % key);
        return;
    }

    bool was_pressed = press_key(key);
    if (_sticky_info.trace_is_on) {
        if (key == _sticky_info.key) {
            _sticky_info.key_down = true;
        }

        if (!_sticky_info.sticky_mode) {
            // during tracing (traced key was pressed and no keyboard event has occured till now)
            if (_sticky_info.key_first_down) {
                ASSERT(_sticky_info.key != REDKEY_INVALID);
                if (key != _sticky_info.key) {
                    reset_sticky();
                }
            } else if (is_sticky_trace_key(key) && (_num_keys_pressed == 1) && !was_pressed) {
                ASSERT(_sticky_info.key == REDKEY_INVALID);
                // start tracing
                _sticky_info.key =  key;
                _sticky_info.key_first_down = true;
                _sticky_info.key_down = true;
                Platform::activate_interval_timer(_sticky_info.timer, STICKY_KEY_TIMEOUT);
            }
        }
    }


    if (_send_cad && _active_screen && _active_screen->intercepts_sys_key() && is_cad_pressed()) {
        send_alt_ctl_del();
        return;
    }

    int action = is_hotkeys_pressed();
    if (action != SpiceConfiguration::ActionUnknown) {
        if (_actions.find(action) != _actions.end()) {
            (this->*_actions[action])();
            return;
        }
    }

#ifdef WIN32
    if (!_active_screen->intercepts_sys_key() && (key == REDKEY_LEFT_CMD ||key == REDKEY_RIGHT_CMD ||
                                        key == REDKEY_MENU || _key_table[REDKEY_L_ALT].press )) {
        unpress_key(key);
        return;
    }
    if (!_sticky_info.sticky_mode &&
        ((_key_table[REDKEY_L_CTRL].press || _key_table[REDKEY_R_CTRL].press) &&
        (_key_table[REDKEY_L_ALT].press || _key_table[REDKEY_R_ALT].press))) {
        if (key == REDKEY_END || key == REDKEY_PAD_1) {
            unpress_key(key);
            _inputs_handler->on_key_down(get_make_scan_code(REDKEY_DELETE));
            _inputs_handler->on_key_up(get_break_scan_code(REDKEY_DELETE));
        } else if (key == REDKEY_DELETE || key == REDKEY_PAD_POINT) {
            unpress_key(key);
            return;
        }
    }
#endif

    _inputs_handler->on_key_down(scan_code);
}

void Application::do_on_key_up(RedKey key)
{
    unpress_key(key);
    uint32_t scan_code = get_break_scan_code(key);
    if (!scan_code) {
        WARN(fmt("no break code for %d") % key);
        return;
    }
    _inputs_handler->on_key_up(scan_code);
}

void Application::on_key_up(RedKey key)
{
    if(key < 0 || key >= REDKEY_NUM_KEYS || !_key_table[key].press) {
        return;
    }

    if (_sticky_info.trace_is_on) {
        ASSERT(_sticky_info.sticky_mode || (key == _sticky_info.key) ||
               (_sticky_info.key == REDKEY_INVALID));
        if (key == _sticky_info.key) {
            _sticky_info.key_down = false;
            if (_sticky_info.key_first_down) {
                _sticky_info.key_first_down = false;
                if (!_sticky_info.sticky_mode) {
                    reset_sticky();
                } else {
                    return; // ignore the sticky-key first release
                }
            }
        }

        if (_sticky_info.sticky_mode) {
            RedKey old_sticky_key = _sticky_info.key;
            reset_sticky();
            if (key == old_sticky_key) {
                return; // no need to send key_up twice
            }
        }
    }

    do_on_key_up(key);
 }

void Application::on_deactivate_screen(RedScreen* screen)
{
    if (_active_screen == screen) {
        _sys_key_intercept_mode = false;
        release_capture();
        _active_screen = NULL;
    }
}

void Application::on_activate_screen(RedScreen* screen)
{
    ASSERT(!_active_screen || (_active_screen == screen));
    _active_screen = screen;
    sync_keyboard_modifiers();
}

void Application::on_start_screen_key_interception(RedScreen* screen)
{
    ASSERT(screen == _active_screen);

    _sys_key_intercept_mode = true;
    reset_sticky();
}

void Application::on_stop_screen_key_interception(RedScreen* screen)
{
    ASSERT(screen == _active_screen);

    _sys_key_intercept_mode = false;
    reset_sticky();
}

void Application::on_app_activated()
{
    _active = true;
    _inputs_handler->on_focus_in();
    _foreign_menu.on_menu_activated();
}

void Application::on_app_deactivated()
{
    _active = false;
    _inputs_handler->on_focus_out();
    _foreign_menu.on_menu_deactivated();
#ifdef WIN32
    if (!_changing_screens) {
        exit_full_screen();
    }
#endif
}

bool Application::rearrange_monitors(RedScreen& screen)
{
    if (!_full_screen) {
        return false;
    }
    bool capture = release_capture();
    prepare_monitors();
    position_screens();
    if (capture && _main_screen != &screen) {
        capture = false;
        _main_screen->activate();
        _main_screen->capture_inputs();
    }
    return capture;
}

Monitor* Application::find_monitor(int id)
{
    ASSERT(_monitors);
    std::list<Monitor*>::const_iterator iter = _monitors->begin();
    for (; iter != _monitors->end(); iter++) {
        Monitor *mon = *iter;
        if (mon->get_id() == id) {
            return mon;
        }
    }
    return NULL;
}

Monitor* Application::get_monitor(int id)
{
    Monitor *mon = find_monitor(id);
    if ((mon = find_monitor(id))) {
        mon->set_used();
    }
    return mon;
}

void Application::assign_monitors()
{
    for (int i = 0; i < (int)_screens.size(); i++) {
        if (_screens[i]) {
            ASSERT(!_screens[i]->has_monitor());
            _screens[i]->set_monitor(get_monitor(i));
        }
    }
}

void Application::prepare_monitors()
{
    for (int i = 0; i < (int)_screens.size(); i++) {
        Monitor* mon;
        if (_screens[i] && (mon = _screens[i]->get_monitor())) {
            Point size = _screens[i]->get_size();
            mon->set_mode(size.x, size.y);
        }
    }
    //todo: test match of monitors size/position against real world size/position
}

void Application::restore_monitors()
{
    //todo: renew monitors (destroy + init)
    std::list<Monitor*>::const_iterator iter = _monitors->begin();
    for (; iter != _monitors->end(); iter++) {
        (*iter)->restore();
    }
}

void Application::position_screens()
{
    for (int i = 0; i < (int)_screens.size(); i++) {
        Monitor* mon;
        if (_screens[i] && (mon = _screens[i]->get_monitor())) {
            _screens[i]->position_full_screen(mon->get_position());
        }
    }
}

void Application::hide()
{
    for (int i = 0; i < (int)_screens.size(); i++) {
        if (_screens[i]) {
            _screens[i]->hide();
        }
    }
}

void Application::show()
{
    for (int i = 0; i < (int)_screens.size(); i++) {
        if (_screens[i]) {
            _screens[i]->show();
        }
    }
}

void Application::external_show()
{
    LOG_DEBUG("Entry, _screens.size()=" << _screens.size());
    for (size_t i = 0; i < _screens.size(); ++i) {
    DBG (0, fmt ("%d") % i);
        if (_screens[i]) {
            _screens[i]->external_show();
        }
    }
}

void Application::show_full_screen()
{
    for (int i = 0; i < (int)_screens.size(); i++) {
        if (_screens[i]) {
            _screens[i]->show_full_screen();
        }
    }
}

void Application::enter_full_screen()
{
    LOG_INFO("");
    _changing_screens = true;
    release_capture();
    assign_monitors();
    hide();
    prepare_monitors();
    position_screens();
    show_full_screen();
    _main_screen->activate();
    _main_screen->capture_inputs();
    _changing_screens = false;
    _full_screen = true;
}

void Application::exit_full_screen()
{
    if (!_full_screen) {
        return;
    }
    LOG_INFO("");
    _changing_screens = true;
    release_capture();
    for (int i = 0; i < (int)_screens.size(); i++) {
        if (_screens[i]) {
             Monitor* mon;
            _screens[i]->exit_full_screen();
            if ((mon = _screens[i]->get_monitor())) {
                _screens[i]->set_monitor(NULL);
                mon->set_free();
            }
        }
    }
    restore_monitors();
    _full_screen = false;
    show();
    _main_screen->activate();
    _changing_screens = false;
}

bool Application::toggle_full_screen()
{
    if (_full_screen) {
        exit_full_screen();
    } else {
        enter_full_screen();
    }
    return _full_screen;
}

void Application::minimize()
{
    for (int i = 0; i < (int)_screens.size(); i++) {
        if (_screens[i]) {
            _screens[i]->minimize();
        }
    }
}

void Application::destroy_monitors()
{
    for (int i = 0; i < (int)_screens.size(); i++) {
        if (_screens[i]) {
            _screens[i]->set_monitor(NULL);
        }
    }
    Platform::destroy_monitors();
    _monitors = NULL;
}

void Application::init_monitors()
{
    exit_full_screen();
    destroy_monitors();
    _monitors = &Platform::init_monitors();
}

void Application::on_monitors_change()
{
    if (Monitor::is_self_change()) {
        return;
    }
    exit_full_screen();
    init_monitors();
}

void Application::on_display_mode_change()
{
    _client.on_display_mode_change();
}

void Application::show_splash(int screen_id)
{
    if (screen_id != 0) {
        return;
    }
    _splash_mode = true;
    release_capture();
    ASSERT(!_sticky_info.trace_is_on);
    (*_gui_layer).set_splash_mode();
}

void Application::hide_splash(int screen_id)
{
    if (screen_id != 0) {
        return;
    }
    _splash_mode = false;
    (*_gui_layer).set_info_mode();
    reset_sticky();
}

void Application::set_peer_connect_options(peer_connect_options configuration[RED_CHANNEL_END])
{
    for (int i=0; i < RED_CHANNEL_END; i++) {
        _peers_configuration[i] = configuration[i];
    }
}

#ifdef WIN32

void Application::set_ipc_interface(CSpiceXIPC* pIPCInterface)
{
    _spice_ipc = pIPCInterface;
    if (_spice_ipc) {
        _spice_ipc->SetOwner(this);
    }
}

#endif // WIN32

void Application::set_controller(SpiceController *controller)
{
    _external_controller.reset(controller);
}

void Application::set_hotkeys(const SpiceConfiguration::hotkeys_t& hotkeys)
{
    _hotkeys = hotkeys;
}

void Application::set_send_cad(bool sendcad)
{
    _send_cad = sendcad;
}

uint32_t Application::get_mouse_mode()
{
    return _client.get_mouse_mode();
}

std::string Application::get_canvas_type()
{
    return _canvas_type;
}


#ifdef WIN32
void Application::set_no_taskmgr(bool no_taskmgr)
{
    _no_taskmgr = no_taskmgr;

    for (size_t i = 0; i < _screens.size(); ++i) {
        if (_screens[i]) {
            _screens[i]->set_no_taskmgr(_no_taskmgr);
        }
    }
}
#endif

void Application::set_title(std::wstring& title)
{
    _title = title;
    update_title();
}

void Application::update_title()
{
    std::wstring cur_title;
    if (get_mouse_mode() == RED_MOUSE_MODE_SERVER) {
        cur_title = _title;
    } else {
        std::wstring edge_str = L"%d";
        size_t edge_pos = _title.find(edge_str);
        if (edge_pos != std::string::npos) {
            edge_pos += edge_str.length();
        }
        cur_title = _title.substr(0, edge_pos);
    }

    for (size_t i = 0; i < _screens.size(); ++i) {
        if (_screens[i]) {
            _screens[i]->set_name(cur_title);
        }
    }
}

#ifdef WIN32

void Application::create_menu(const wchar_t *resource)
{
    _menu_resource = resource;

    for (size_t i = 0; i < _screens.size(); ++i) {
        if (_screens[i]) {
            _screens[i]->create_menu(_menu_resource.c_str());
        }
    }
}

void Application::delete_menu()
{
    for (size_t i = 0; i < _screens.size(); ++i) {
        if (_screens[i]) {
            _screens[i]->delete_menu();
        }
    }
}

void Application::menu_item_callback(unsigned int item_id)
{
    DBG(0, "menu_item_callback>>> item " << item_id << "signaled");

    if (_spice_ipc) {
        _spice_ipc->NotifyMenuItemClick(item_id);
    }
}

#endif

int Application::is_hotkeys_pressed()
{
    SpiceConfiguration::hotkeys_t::const_iterator iter = _hotkeys.begin();

    while (iter != _hotkeys.end()) {
        if (is_action_keys_pressed(iter->second)) {
            break;
        }
        ++iter;
    }

    return ((iter != _hotkeys.end()) ? iter->first : SpiceConfiguration::ActionUnknown);
}

bool Application::is_action_keys_pressed(const SpiceConfiguration::hotkey_seq_t& keys)
{
    SpiceConfiguration::hotkey_seq_t::const_iterator iter = keys.begin();

    while (iter != keys.end()) {
        if (!(_key_table[iter->main].press || _key_table[iter->alter].press)) {
            break;
        }
        ++iter;
    }

    return (iter == keys.end());
}

bool Application::is_cad_pressed()
{
    return ((_key_table[REDKEY_L_CTRL].press || _key_table[REDKEY_R_CTRL].press) &&
            (_key_table[REDKEY_L_ALT].press || _key_table[REDKEY_R_ALT].press) &&
            (_key_table[REDKEY_DELETE].press || _key_table[REDKEY_PAD_POINT].press));
}

void Application::send_key_down(RedKey key)
{
    _inputs_handler->on_key_down(get_make_scan_code(key));
}

void Application::send_key_up(RedKey key)
{
    _inputs_handler->on_key_up(get_break_scan_code(key));
}

void Application::send_alt_ctl_del()
{
    send_key_down(REDKEY_L_CTRL);
    send_key_down(REDKEY_L_ALT);
    send_key_down(REDKEY_DELETE);

    send_key_up(REDKEY_DELETE);
    send_key_up(REDKEY_L_ALT);
    send_key_up(REDKEY_L_CTRL);
}

void Application::send_ctrl_alt_end()
{
    send_key_down(REDKEY_L_CTRL);
    send_key_down(REDKEY_L_ALT);
    send_key_down(REDKEY_END);

    send_key_up(REDKEY_L_CTRL);
    send_key_up(REDKEY_L_ALT);
    send_key_up(REDKEY_END);
}

void Application::send_action_hotkeys(int action)
{
    SpiceConfiguration::hotkeys_t::const_iterator iter = _hotkeys.find(action);
    if (iter != _hotkeys.end()) {
        send_hotkeys(iter->second);
    }
}

void Application::send_hotkeys(const SpiceConfiguration::hotkey_seq_t& hotkeys)
{
    SpiceConfiguration::hotkey_seq_t::const_iterator iter;

    for (iter = hotkeys.begin(); iter != hotkeys.end(); ++iter) {
        send_key_down(iter->main);
    }

    for (iter = hotkeys.begin(); iter != hotkeys.end(); ++iter) {
        send_key_up(iter->main);
    }
}


void Application::config(SpiceConfiguration& spice_config)
{
    LOG_TRACE("");

    if (spice_config._init_controller) {
        // under linux, config will be called again, but with _init_controller=false
        _controller_is_up = true;
#ifndef WIN32
        _external_controller.reset(new SpiceController(&_config));
        _external_controller->SetApp(this);
        Platform::init_from_ipc(_external_controller.get());
        return;
#endif
    }
    set_ssl_config(spice_config);

    if(!spice_config._no_audio) {
        _client.register_channel_factory(PlaybackChannel::Factory());
        _client.register_channel_factory(RecordChannel::Factory());
    }

    set_title (spice_config._title);

    set_peer_connect_options(_peers_configuration);
    set_hotkeys(spice_config._hotkeys);
    set_send_cad(spice_config._send_cad);
#ifdef WIN32
    set_no_taskmgr(spice_config._no_taskmgr);
#endif
    _canvas_type = spice_config._canvas_type;
    _client.init(spice_config._host.c_str(), spice_config._port,spice_config._password.c_str(),
                 spice_config._auto_display_res);
    if (_client.is_auto_display_res()) {
        Monitor* mon = find_monitor(0);
        ASSERT(mon);
        Point size = mon->get_size();
        _main_screen->set_mode(size.x, size.y, 32);
    }

    if (spice_config._full_screen) {
        enter_full_screen();
    } else {
        _main_screen->show(true, NULL);
    }

#ifdef WIN32
    if (_external_controller.get()) {
        _external_controller->SetApp(this);
    }
#endif
}
