#include "common.h"

#include <string>

#include "debug.h"
#include "utils.h"
#include "foreign_menu.h"
#include "foreign_menu_prot.h"
#include "platform.h"

ForeignMenu::ForeignMenu(ForeignMenuInterface *handler)
    : _handler (handler)
    , _active (false)
{
    ASSERT(_handler != NULL);
#ifdef WIN32
    const std::string& conn_name = (fmt("SpiceForeignMenu-%d") % ::GetCurrentProcessId()).str();
#else
    const std::string& conn_name = (fmt("/tmp/SpiceForeignMenu-%d.uds") % ::getpid()).str();
#endif
    LOG_INFO("Creating a foreign menu connection: " << conn_name);
    _foreign_menu = NamedPipe::create(conn_name.c_str(), *this);
    if (!_foreign_menu) {
        LOG_ERROR("Failed to create a foreign menu connection.");
    }
}

ForeignMenu::~ForeignMenu()
{
    if (_foreign_menu) {
        NamedPipe::destroy(_foreign_menu);
    }
}

NamedPipe::ConnectionInterface& ForeignMenu::create()
{
    ForeignMenuItem *menu_item = new ForeignMenuItem(_handler, *this);

    if (menu_item == NULL) {
        throw Exception("Error allocating a new foreign menu item.");
    }
    return *menu_item;
}

void ForeignMenu::add_item_connection(NamedPipe::ConnectionRef conn_ref, ForeignMenuItem *menu_item)
{
    _connections[conn_ref] = menu_item;
    if (_active) {
        handle_menu_activated(menu_item, FOREIGN_MENU_ACTIVATE);
    }
    while (menu_item->read_msgs());
}

void ForeignMenu::on_menu_activated()
{
    std::map<NamedPipe::ConnectionRef, ForeignMenuItem*>::const_iterator conn;
    _active = true;
    for (conn = _connections.begin(); conn != _connections.end(); ++conn) {
        handle_menu_activated(conn->second, FOREIGN_MENU_ACTIVATE);
    }
}

void ForeignMenu::on_menu_deactivated()
{
    std::map<NamedPipe::ConnectionRef, ForeignMenuItem*>::const_iterator conn;
    _active = false;
    for (conn = _connections.begin(); conn != _connections.end(); ++conn) {
        handle_menu_activated(conn->second, FOREIGN_MENU_DEACTIVATE);
    }
}

void ForeignMenu::on_foreign_menu_command(NamedPipe::ConnectionRef conn_ref, int32_t id)
{
    FrgMenuEvent msg;

    msg.base.magic = FOREIGN_MENU_MAGIC;
    msg.base.id = FOREIGN_MENU_EVENT;
    msg.base.size = sizeof(FrgMenuEvent);
    msg.id = id;
    msg.action = FOREIGN_MENU_EVENT_CLICK;
    send_msg_to_client(conn_ref, &msg.base);
}

void ForeignMenu::item_connection_closed(NamedPipe::ConnectionRef conn_ref)
{
    ForeignMenuItem *item = _connections[conn_ref];
    _connections.erase(conn_ref);
    delete item;
}

void ForeignMenu::handle_menu_activated(ForeignMenuItem *item, int32_t cmd)
{
    FrgMenuMsg msg;

    msg.magic = FOREIGN_MENU_MAGIC;
    msg.id = cmd;
    msg.size = sizeof(FrgMenuMsg);
    send_msg_to_client(item, &msg);
}

void ForeignMenu::send_msg_to_client(NamedPipe::ConnectionRef conn_ref, FrgMenuMsg *msg)
{
    send_msg_to_client(_connections[conn_ref], msg);
}

void ForeignMenu::send_msg_to_client(ForeignMenuItem *item, FrgMenuMsg *msg)
{
    ASSERT(item != NULL);
    item->write_msg(msg, msg->size);
}

ForeignMenuItem::ForeignMenuItem(ForeignMenuInterface *handler, ForeignMenu& parent)
    : _handler (handler)
    , _parent (parent)
    , _write_pending (0)
    , _write_pos (_write_buf)
    , _read_pos (_read_buf)
{
}

ForeignMenuItem::~ForeignMenuItem()
{
    if (_opaque != NamedPipe::INVALID_CONNECTION) {
        _handler->delete_menu(_opaque);
        NamedPipe::destroy_connection(_opaque);
    }
}

void ForeignMenuItem::bind(NamedPipe::ConnectionRef conn_ref)
{
    _opaque = conn_ref;
    _parent.add_item_connection(conn_ref, this);
}

void ForeignMenuItem::on_data()
{
    if (_write_pending) {
        LOG_INFO("Resume pending write " << _write_pending);
        write_msg(_write_pos, _write_pending);
    }
    while (read_msgs());
}

bool ForeignMenuItem::read_msgs()
{
    uint8_t *head_ptr = _read_buf;
    size_t nread = _read_pos - _read_buf;
    int32_t size;

    ASSERT(_opaque != NamedPipe::INVALID_CONNECTION);
    size = NamedPipe::read(_opaque, (uint8_t*)_read_pos, sizeof(_read_buf) - nread);
    if (size == 0) {
        return false;
    } else if (size < 0) {
        LOG_DEBUG("Error reading named pipe data (size = " << size << ")");
        _parent.item_connection_closed(_opaque);
        return false;
    }
    nread += size;
    while (nread > 0) {
        if (nread < sizeof(FrgMenuMsg)) {
            break;
        }
        FrgMenuMsg *hdr = (FrgMenuMsg *)head_ptr;
        if (hdr->magic != FOREIGN_MENU_MAGIC || hdr->size == 0) {
            LOG_DEBUG("Bad foreign menu message (magic = " << hdr->magic << ", size = " << hdr->size << ")");
            _parent.item_connection_closed(_opaque);
            return false;
        }
        if (nread < hdr->size) {
            break;
        }
        handle_foreign_menu_message(hdr, _opaque);
        nread -= hdr->size;
        head_ptr += hdr->size;
    }
    if (nread > 0 && head_ptr != _read_buf) {
        memcpy(_read_buf, head_ptr, nread);
    }
    _read_pos = _read_buf + nread;
    return true;
}

void ForeignMenuItem::write_msg(const void *buf, int len)
{
    uint8_t *pos;
    int32_t written = 0;

    ASSERT(_opaque != NamedPipe::INVALID_CONNECTION);
    if (_write_pending && _write_buf != buf) {
        LOG_DEBUG("Dropping msg due to pending write " << _write_pending);
        return;
    }
    pos = (uint8_t*)buf;  
    while (len && (written = NamedPipe::write(_opaque, pos, len)) > 0) {
        pos += written;
        len -= written;
    }
    if (len && written == 0) {
        if (_write_pending) {
            _write_pos = pos;
        } else {
            _write_pos = _write_buf;
            memcpy(_write_buf, pos, len);
        }
        _write_pending = len;
    } else if (written < 0) {
        LOG_DEBUG("Error writing to named pipe " << written);
        _parent.item_connection_closed(_opaque);
    } else {
        _write_pending = 0;
    }
}
void ForeignMenuItem::handle_foreign_menu_message(FrgMenuMsg *hdr, NamedPipe::ConnectionRef conn_ref)
{
    ASSERT(_handler != NULL);
    _handler->on_foreign_menu_message(hdr, conn_ref);
}
