#include "common.h"
#ifdef _WIN32
#include <winsock2.h>
#include <ws2tcpip.h>

#define SHUT_RDWR SD_BOTH
#else
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/tcp.h>
#include <netdb.h>

#define INVALID_SOCKET -1
#define SOCKET_ERROR -1
#define closesocket(sock) ::close(sock)
#endif
#include <openssl/x509.h>
#include <openssl/x509v3.h>
#include "red.h"
#include "redpeer.h"
#include "utils.h"
#include "debug.h"

#ifdef _WIN32

int inet_aton(const char *ip, struct in_addr *in_addr)
{
    unsigned long addr = inet_addr(ip);

    if (addr == INADDR_NONE) {
        return 0;
    }
    in_addr->S_un.S_addr = addr;//htonl(addr);
    return 1;
}

#define SHUTDOWN_ERR WSAESHUTDOWN
#define INTERRUPTED_ERR WSAEINTR
#define WOULDBLOCK_ERR WSAEWOULDBLOCK
#define sock_error() WSAGetLastError()
#define sock_err_message() WSAGetLastError()
#else
#define SHUTDOWN_ERR EPIPE
#define INTERRUPTED_ERR EINTR
#define WOULDBLOCK_ERR EAGAIN
#define sock_error() errno
#define sock_err_message() strerror(errno)
#endif

#if 0
// this is used only if client certificate code is enabled.
static int ssl_password_cb (
	char *buf,
	int nSize,
	int nFlags,
	void *data
) {
	char *pass = (char *)data;
	
	if (nSize < (int)strlen (pass) +1) {
		return (0);
	}
	
	strcpy (buf,pass);
	return(strlen(pass));
}
#endif 

typedef struct SslVerifyCbData {
    peer_connect_options info;
    const char* host_name;
    bool all_preverify_ok;
} SslVerifyCbData;

static void ssl_error()
{
    unsigned long last_error = ERR_peek_last_error();

    ERR_print_errors_fp(stderr);
    throw Exception(fmt("%s: SSL Error %s") % FUNC_NAME % ERR_error_string(last_error, NULL),
                    SPICEC_ERROR_CODE_SSL_ERROR);
}

RedPeer::RedPeer()
    : _peer (INVALID_SOCKET)
    , _shut (false)
{
}

RedPeer::~RedPeer()
{
    if (_peer != INVALID_SOCKET) {
        closesocket(_peer);
    }
}

bool RedPeer::verify_pubkey(X509* cert, const uint8_t* key, int key_len)
{
    EVP_PKEY* cert_pubkey = NULL;
    EVP_PKEY* orig_pubkey = NULL;
    BIO* bio = NULL;
    int ret = 0;

    if (!key_len) {
        return false;
    }

    ASSERT(key);
    ASSERT(cert);

    try {
        cert_pubkey = X509_get_pubkey(cert);
        if (!cert_pubkey) {
            throw Exception("reading public key from certificate failed");
        }

        bio = BIO_new_mem_buf((void*)key, key_len);
        if (!bio) {
            throw Exception("creating BIO failed");
        }

        orig_pubkey = d2i_PUBKEY_bio(bio, NULL);
        if (!orig_pubkey) {
            throw Exception("reading pubkey from bio failed");
        }

        ret = EVP_PKEY_cmp(orig_pubkey, cert_pubkey);
     
        BIO_free(bio);
        EVP_PKEY_free(orig_pubkey);
        EVP_PKEY_free(cert_pubkey);
        if (ret == 1) {
            DBG(0, "public keys match");
            return true;
        } else if (ret == 0) {
            DBG(0, "public keys mismatch");
            return false;
        } else {
            DBG(0, "public keys types mismatch");
            return false;
        }
    } catch(Exception& e) {
        LOG_WARN(e.what());
        if (bio) {
            BIO_free(bio);
        }

        if (orig_pubkey) {
            EVP_PKEY_free(orig_pubkey);
        }

        if (cert_pubkey) {
            EVP_PKEY_free(cert_pubkey);
        }
        return false;
    }
}

/* From gnutls: compare host_name against certificate, taking account of wildcards.
 * return true on success or false on error.
 *
 * note: cert_name_size is required as X509 certs can contain embedded NULs in
 * the strings such as CN or subjectAltName
 */
bool RedPeer::x509_cert_host_name_compare(const char *cert_name, int cert_name_size,
                                          const char *host_name)
{
    /* find the first different character */
    for (; *cert_name && *host_name && (toupper(*cert_name) == toupper(*host_name));
         cert_name++, host_name++, cert_name_size--);

    /* the strings are the same */
    if (cert_name_size == 0 && *host_name == '\0')
        return true;

    if (*cert_name == '*')
    {
        /* a wildcard certificate */
        cert_name++;
        cert_name_size--;

        while (true)
        {
            /* Use a recursive call to allow multiple wildcards */
            if (RedPeer::x509_cert_host_name_compare(cert_name, cert_name_size, host_name)) {
                return true;
            }

	        /* wildcards are only allowed to match a single domain
               component or component fragment */
            if (*host_name == '\0' || *host_name == '.')
                break;
            host_name++;
	    }

        return false;
    }

  return false;
}

/*
 * From gnutls_x509_crt_check_hostname - compares the hostname with certificate's hostname
 *
 * This function will check if the given certificate's subject matches
 * the hostname.  This is a basic implementation of the matching
 * described in RFC2818 (HTTPS), which takes into account wildcards,
 * and the DNSName/IPAddress subject alternative name PKIX extension.
 *
 */
bool RedPeer::verify_host_name(X509* cert, const char* host_name)
{
    GENERAL_NAMES* subject_alt_names;
    bool found_dns_name = false;
    struct in_addr addr;
    int addr_len = 0;
    bool cn_match = false;

    ASSERT(cert);

    // only IpV4 supported
    if (inet_aton(host_name, &addr)) {
        addr_len = sizeof(struct in_addr);
    }

    /* try matching against:
    *  1) a DNS name or IP address as an alternative name (subjectAltName) extension
    *     in the certificate
    *  2) the common name (CN) in the certificate
    *
    *  either of these may be of the form: *.domain.tld
    *
    *  only try (2) if there is no subjectAltName extension of
    *  type dNSName
    */


    subject_alt_names = (GENERAL_NAMES*)X509_get_ext_d2i(cert, NID_subject_alt_name, NULL, NULL);

    if (subject_alt_names) {
        int num_alts = sk_GENERAL_NAME_num(subject_alt_names);
        for (int i = 0; i < num_alts; i++) {
            const GENERAL_NAME* name = sk_GENERAL_NAME_value(subject_alt_names, i);
            if (name->type == GEN_DNS) {
                found_dns_name = true;
                if (RedPeer::x509_cert_host_name_compare((char *)ASN1_STRING_data(name->d.dNSName),
                                                         ASN1_STRING_length(name->d.dNSName),
                                                         host_name)) {
                    DBG(0, fmt("alt name match=%s") % ASN1_STRING_data(name->d.dNSName));
                    GENERAL_NAMES_free(subject_alt_names);
                    return true;
                }
            } else if (name->type == GEN_IPADD) {
                int alt_ip_len = ASN1_STRING_length(name->d.iPAddress);
                found_dns_name = true;
                if ((addr_len == alt_ip_len)&& 
                    !memcmp(ASN1_STRING_data(name->d.iPAddress), &addr, addr_len)) {
                         DBG(0, fmt("alt name IP match=%s") % 
                                inet_ntoa(*((struct in_addr*)ASN1_STRING_data(name->d.dNSName))));
                    GENERAL_NAMES_free(subject_alt_names);
                    return true;
                }
            }
        }
        GENERAL_NAMES_free(subject_alt_names);
    }

    if (found_dns_name)
    {
        DBG(0, "SubjectAltName mismatch");
        return false;
    }

    /* extracting commonNames */
    X509_NAME* subject = X509_get_subject_name(cert);
    if (subject) {
        int pos = -1;
        X509_NAME_ENTRY* cn_entry;
        ASN1_STRING* cn_asn1;
        
        while ((pos = X509_NAME_get_index_by_NID(subject, NID_commonName, pos)) != -1) {
            cn_entry = X509_NAME_get_entry(subject, pos);
            if (!cn_entry) {
                continue;
            }
            cn_asn1 = X509_NAME_ENTRY_get_data(cn_entry);
            if (!cn_asn1) {
                continue;
            }

            if (RedPeer::x509_cert_host_name_compare((char*)ASN1_STRING_data(cn_asn1),
                                                     ASN1_STRING_length(cn_asn1),
                                                     host_name)) {
                DBG(0, fmt("common name match=%s") % (char*)ASN1_STRING_data(cn_asn1));
                cn_match = true;
                break;
            }         
        }
    }
    
    if (!cn_match) {
        DBG(0, "common name mismatch");
    }
    return cn_match;    

}

bool RedPeer::verify_subject(X509* cert, const CertFieldValueList& subject)
{
    X509_NAME* cert_subject = NULL;
    CertFieldValueList::const_iterator subject_iter;
    X509_NAME* in_subject;
    int ret;

    ASSERT(cert);

    cert_subject = X509_get_subject_name(cert);
    if (!cert_subject) {
        LOG_WARN("reading certificate subject failed");
        return false;
    } 

    if (X509_NAME_entry_count(cert_subject) != subject.size()) {
        DBG(0, fmt("subject mismatch: #entries cert=%d, input=%d") %
            X509_NAME_entry_count(cert_subject) % subject.size());
        return false;
    }

    in_subject = X509_NAME_new();
    if (!in_subject) {
        LOG_WARN("failed to allocate X509_NAME");
        return false;
    }

    for (subject_iter = subject.begin(); subject_iter != subject.end(); subject_iter++) {
        if (!X509_NAME_add_entry_by_txt(in_subject,
                                        subject_iter->first.c_str(),
                                        MBSTRING_UTF8, 
                                        (const unsigned char*)subject_iter->second.c_str(), 
                                        subject_iter->second.length(), -1, 0)) {
            LOG_WARN(fmt("failed to add entry %s=%s to X509_NAME") % 
                         subject_iter->first.c_str() % subject_iter->second.c_str());
             X509_NAME_free(in_subject);
             return false;
            
        }
    }

    ret = X509_NAME_cmp(cert_subject, in_subject);
    X509_NAME_free(in_subject);
 
    if (ret == 0) {
         DBG(0, "subjects match");
         return true;
    } else {
         DBG(0, "subjects mismatch");
         return false;
    }
}

int RedPeer::ssl_verify_callback(int preverify_ok, X509_STORE_CTX *ctx) 
{
    int depth;
    SSL *ssl;
    X509* cert;
    SslVerifyCbData* verify_data;
    int server_auth_flags;

    depth = X509_STORE_CTX_get_error_depth(ctx);

    ssl = (SSL*)X509_STORE_CTX_get_ex_data(ctx, SSL_get_ex_data_X509_STORE_CTX_idx());
    if (!ssl) {
        LOG_WARN("failed to get ssl connection");
        return 0;
    }

    verify_data = (SslVerifyCbData*)SSL_get_app_data(ssl);
    server_auth_flags = verify_data->info.GetServerAuthFlags();

    if (depth > 0) {
        // if certificate verification failed, we can still authorize the server
        // if its public key matches the one we hold in the peer_connect_options. 
        if (!preverify_ok) {
            DBG(0, fmt("openssl verify failed at depth=%d")%depth);
            verify_data->all_preverify_ok = false;
            if (server_auth_flags & SERVER_AUTH_PUBKEY) {
                return 1;        
            } else {
                return 0;
            }
        } else {
            return preverify_ok;
        }
    }

    /* depth == 0 */
    cert = X509_STORE_CTX_get_current_cert(ctx);
    if (!cert) {
        LOG_WARN("failed to get server certificate");
        return 0;
    }
    
    if (server_auth_flags & SERVER_AUTH_PUBKEY) {
        if (verify_pubkey(cert, verify_data->info.GetPubKey(), 
                          verify_data->info.GetPubKeySize())) {
            return 1;
        }
    }

    if (!verify_data->all_preverify_ok || !preverify_ok) {
        return 0;
    }

    if (server_auth_flags & SERVER_AUTH_NAME) {
        if (verify_host_name(cert, verify_data->host_name)) {
            return 1;
        }
    }

    if (server_auth_flags & SERVER_AUTH_SUBJECT) {
        if (verify_subject(cert, verify_data->info.GetSubject())) {
            return 1;
        }
    }
    return 0;
}

void RedPeer::connect(peer_connect_options& configuration, const char* host, uint16_t port)
{
    struct sockaddr_in addr;
    int no_delay;
    int peer;
    int server_auth_flags;
    SslVerifyCbData server_auth_data;
    uint32_t ip = host_by_name(host);

	if (!configuration.IsSSL()) {
		_ctx = NULL;
	} else {
		SSL_METHOD *ssl_method;
		int return_code;
		
		ssl_method=TLSv1_method();
		_ctx = SSL_CTX_new(ssl_method);
		if (_ctx == NULL) {
			//ERROR
			ssl_error();
		}

#if 0		
		// Code for client authentication is currently disabled in code.
		nRC = SSL_CTX_use_certificate_chain_file (_ctx,rccConf.GetCertFile());
		if (nRC != 1) {
			//ERROR
			ssl_error();
		}

		SSL_CTX_set_default_passwd_cb_userdata (_ctx, (void *)rccConf.GetPassword());	
		SSL_CTX_set_default_passwd_cb (_ctx, ssl_password_cb);
		
		nRC = SSL_CTX_use_PrivateKey_file (_ctx, rccConf.GetKeyFile(), SSL_FILETYPE_PEM);
		if (nRC != 1) {
			//ERROR
			ssl_error();
		}

#endif	
        server_auth_data.info.assign(configuration);
        server_auth_data.host_name = host;
        server_auth_data.all_preverify_ok = true;
        server_auth_flags = configuration.GetServerAuthFlags();
        
        if ((server_auth_flags & SERVER_AUTH_NAME) || (server_auth_flags & SERVER_AUTH_SUBJECT)) {
            DBG(0, fmt("truststore file=%s") % configuration.GetCAFile());
            return_code = SSL_CTX_load_verify_locations (_ctx, configuration.GetCAFile(), NULL);
            if (return_code != 1) {
                if (server_auth_flags & SERVER_AUTH_PUBKEY) {         
                    CertFieldValueList empty_subject;
                    LOG_WARN(fmt("SSL_CTX_load_verify_locations failed, CA_file=%s. "
                             "only pubkey authentication is active") %
                             configuration.GetCAFile());
                    server_auth_data.info.SetSSLServerAuthParameters(SERVER_AUTH_PUBKEY,
                                                                     configuration.GetPubKey(),
                                                                     configuration.GetPubKeySize(),
                                                                     NULL,
                                                                     empty_subject);
                } else {
                    LOG_WARN(fmt("SSL_CTX_load_verify_locations failed CA_file=%s") %
                             configuration.GetCAFile());
                    ssl_error();
                }
            }
        }

        if (server_auth_flags) {
            SSL_CTX_set_verify(_ctx, SSL_VERIFY_PEER, ssl_verify_callback);
        }

        return_code = SSL_CTX_set_cipher_list (_ctx, configuration.GetCipher());
        if (return_code != 1) {
            ssl_error();
        }
    } 

    ASSERT(_peer == INVALID_SOCKET);

    addr.sin_port = htons(configuration.GetPort());
    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = htonl(ip);

    if ((peer = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) == INVALID_SOCKET) {
        throw Exception(fmt("%s: failed to create socket: %||") % __FUNCTION__ % sock_err_message(),
                        SPICEC_ERROR_CODE_SOCKET_FAILED);
    }

    no_delay = 1;
    if (setsockopt(peer, IPPROTO_TCP, TCP_NODELAY, (const char*)&no_delay,sizeof(no_delay)) ==
                                                                                    SOCKET_ERROR) {
        WARN("set TCP_NODELAY failed");
    }

    LOG_INFO(fmt("Connecting %s %d") % inet_ntoa(addr.sin_addr) % configuration.GetPort());

	if (::connect(peer, (struct sockaddr *)&addr, sizeof(sockaddr_in)) == SOCKET_ERROR) {
		closesocket(peer);
        throw Exception(fmt("%s: failed to connect: %||") % __FUNCTION__ % sock_err_message(),
						SPICEC_ERROR_CODE_CONNECT_FAILED);
	}

    if (_ctx) {
		_ssl = SSL_new (_ctx);
        SSL_set_app_data(_ssl, &server_auth_data);
		_sbio = BIO_new_socket (peer, BIO_NOCLOSE);
		SSL_set_bio (_ssl, _sbio, _sbio);
		int return_code = SSL_connect (_ssl);
		if (return_code <= 0) {
            int ssl_error_code = SSL_get_error(_ssl,return_code);
            LOG_WARN(fmt("%s: failed to connect w/SSL, ssl_error %d") % __FUNCTION__ % ssl_error_code);
            closesocket(peer);
            ssl_error();
		}
    }
			
    Lock lock(_lock);
    if (_shut) {
        closesocket(peer);
        throw Exception(fmt("%s: failed to close socket: %||") % FUNC_NAME % sock_err_message(),
                        SPICEC_ERROR_CODE_CONNECT_FAILED);
    }
    _peer = peer;
    _serial = 0;
}

uint32_t RedPeer::host_by_name(const char* host)
{
    struct addrinfo *result = NULL;
    struct sockaddr_in *addr;
    uint32_t return_value;
    int rc;
    
    rc = getaddrinfo (host, NULL, NULL, &result);
    if(rc != 0  || result == NULL) {
        throw Exception(fmt("%s: cannot resolve host address %s") % __FUNCTION__ % host,
                            SPICEC_ERROR_CODE_GETHOSTBYNAME_FAILED);
    }
    
    addr = (sockaddr_in *) result->ai_addr;
    return_value = addr->sin_addr.s_addr;
    
    freeaddrinfo (result);

    LOG_TRACE(host << " = " << return_value);
    return ntohl(return_value);
}

void RedPeer::disconnect()
{
    Lock lock(_lock);

    if (_peer != INVALID_SOCKET) {
		if (_ctx) {
			SSL_shutdown(_ssl);
		}
		
        shutdown(_peer, SHUT_RDWR);
    }
    _shut = true;
}

void RedPeer::close()
{
    Lock lock(_lock);
    if (_peer != INVALID_SOCKET) {
        if (_ctx) {
            SSL_free (_ssl);
            SSL_CTX_free(_ctx);
        }
	
        closesocket(_peer);
        _peer = INVALID_SOCKET;
    }
}

void RedPeer::swap(RedPeer* other)
{
    Lock lock(_lock);
    SOCKET temp_peer = _peer;
    SSL_CTX *temp_ctx = _ctx;
    SSL *temp_ssl = _ssl;
    BIO *temp_sbio = _sbio;

    _peer = other->_peer;
    other->_peer = temp_peer;

	if (_ctx) {
		_ctx = other->_ctx;
		_ssl = other->_ssl;
		_sbio = other->_sbio;
	    
		other->_ctx = temp_ctx;
		other->_ssl = temp_ssl;
		other->_sbio = temp_sbio;
	}
	    
    if (_shut) {
        shutdown(_peer, SHUT_RDWR);
    }
}

uint32_t RedPeer::recive(uint8_t *buf, uint32_t size)
{
    uint8_t *pos = buf;
    while (size) {
        int now;
        if (_ctx == NULL) {
            if ((now = recv(_peer, (char *)pos, size, 0)) <= 0) {
                if (now == SOCKET_ERROR && sock_error() == WOULDBLOCK_ERR) {
                    break;
                }
                
                if (now == 0 || sock_error() == SHUTDOWN_ERR) {
                    throw RedPeer::DisconnectedException();
                }

                if (sock_error() == INTERRUPTED_ERR) {
                     continue;
                }

                throw Exception(fmt("%s: %||") % __FUNCTION__ % sock_err_message(),
                                SPICEC_ERROR_CODE_RECV_FAILED);
            }
            size -= now;
            pos += now;
        } else {
            if ((now = SSL_read(_ssl, pos, size)) <= 0) {
                int ssl_error = SSL_get_error(_ssl, now);

                if (ssl_error == SSL_ERROR_WANT_READ ) {
                     break;
                }

                if (ssl_error == SSL_ERROR_SYSCALL) {
                    if (now == -1) {
                        if (sock_error() == WOULDBLOCK_ERR) {
                            break;
                        }
                        if (sock_error() == INTERRUPTED_ERR) {
                            continue;
                        }
                    }
                    if (now == 0 || (now == -1 && sock_error() == SHUTDOWN_ERR)) {
                        throw RedPeer::DisconnectedException();
                    }
                } else if (ssl_error == SSL_ERROR_ZERO_RETURN) {
                    throw RedPeer::DisconnectedException();
                }

                throw Exception(fmt("%s: %||") % __FUNCTION__ % sock_err_message(),
                    SPICEC_ERROR_CODE_RECV_FAILED);
            }
            size -= now;
            pos += now;
        }
    }
    return pos - buf;
}

RedPeer::CompundInMessage* RedPeer::recive()
{
    RedDataHeader header;
    std::auto_ptr<CompundInMessage> message;

    recive((uint8_t*)&header, sizeof(RedDataHeader));
    message.reset(new CompundInMessage(header.serial, header.type, header.size, header.sub_list));
    recive(message->data(), message->size());
    return message.release(); 
}

uint32_t RedPeer::send(uint8_t *buf, uint32_t size)
{
    uint8_t *pos = buf;
    while (size) {
        int now;

        if (_ctx == NULL) {
            if ((now = ::send(_peer, (char *)pos, size, 0)) == SOCKET_ERROR) {
                if (sock_error() == WOULDBLOCK_ERR) {
                    break;
                }
                if (sock_error() == SHUTDOWN_ERR) {
                    throw RedPeer::DisconnectedException();
                }
                if (sock_error() == INTERRUPTED_ERR) {
                    continue;
                }
                throw Exception(fmt("%s: %||") % __FUNCTION__ % sock_err_message(),
                                SPICEC_ERROR_CODE_SEND_FAILED);
            }
            size -= now;
            pos += now;
        } else {
            if ((now = SSL_write(_ssl, pos, size)) <= 0) {
                int ssl_error = SSL_get_error(_ssl, now);

                if (ssl_error == SSL_ERROR_WANT_WRITE ) {
                     break;
                }

                if (ssl_error == SSL_ERROR_SYSCALL) {
                    if (now == -1) {
                        if (sock_error() == WOULDBLOCK_ERR) {
                            break;
                        }
                        if (sock_error() == INTERRUPTED_ERR) {
                            continue;
                        }
                    }
                    if (now == 0 || (now == -1 && sock_error() == SHUTDOWN_ERR)) {
                        throw RedPeer::DisconnectedException();
                    }
                } else if (ssl_error == SSL_ERROR_ZERO_RETURN) {
                    throw RedPeer::DisconnectedException();
                }

                throw Exception(fmt("%s: %||") % __FUNCTION__ % sock_err_message(),
                                SPICEC_ERROR_CODE_SEND_FAILED);
            }
            size -= now;
            pos += now;
        }
    }
    return pos - buf;
}

uint32_t RedPeer::send(RedPeer::OutMessage& message)
{
    message.header().serial = ++_serial;
    return send(message.base(), message.message_size());
}



RedPeer::OutMessage::OutMessage(uint32_t type, uint32_t size)
    : _data (new uint8_t[size + sizeof(RedDataHeader)])
    , _size (size)
{
    header().type = type;
    header().size = size;
}

RedPeer::OutMessage::~OutMessage()
{ 
    delete[] _data;
}

void RedPeer::OutMessage::resize(uint32_t size)
{
    if (size <= _size) {
        header().size = size;
        return;
    } 
    uint32_t type = header().type;
    delete[] _data;
    _data = NULL;
    _size = 0;
    _data = new uint8_t[size + sizeof(RedDataHeader)];
    _size = size;
    header().type = type;
    header().size = size;
}

