#ifndef _H_UTILS
#define _H_UTILS

#include "rederrorcodes.h"
#include "common.h"

#define MIN(x, y) (((x) <= (y)) ? (x) : (y))
#define MAX(x, y) (((x) >= (y)) ? (x) : (y))
#define ALIGN(a, b) (((a) + ((b) - 1)) & ~((b) - 1))

#ifdef __GNUC__
#ifdef __i386__
#define mb() __asm__ __volatile__ ("lock; addl $0,0(%%esp)": : :"memory")
#else
#define mb() __asm__ __volatile__ ("lock; addl $0,0(%%rsp)": : :"memory")
#endif
#else
#define mb() __asm {lock add [esp], 0}
#endif

class Exception : public std::exception {
public:
    Exception(const std::string& str, int error_code) : _mess(str), _erro_code(error_code) {}
    Exception(const std::string& str) : _mess(str) { _erro_code = SPICEC_ERROR_CODE_ERROR;}
    Exception(boost::format& format, int error_code) : _mess(format.str()), _erro_code(error_code) {}
    Exception(boost::format& format) : _mess(format.str()) {_erro_code = SPICEC_ERROR_CODE_ERROR;}
    virtual ~Exception() throw() {}
    virtual const char* what() const throw() {return _mess.c_str();}
    int get_error_code() {return _erro_code;}

private:
    std::string _mess;
    int _erro_code;
};

template <class T>
class AutoRef {
public:
    AutoRef(T* obj) : _obj(obj) {}
    ~AutoRef() { if (_obj) _obj->unref();}
    T* operator * () {return _obj;}
private:
    T* _obj;
};

template<class T>
class FreeObject {
public:
    void operator() (T p) { delete p;}
};

template<class T, class FreeRes = FreeObject<T>, T invalid = NULL >
class AutoRes {
public:
    AutoRes() : res (invalid) {}
    AutoRes(T inRes) : res (inRes) {}
    ~AutoRes() { set(invalid); }

    void set(T inRes) {if (res != invalid) free_res(res); res = inRes; }
    T get() {return res;}
    T release() {T tmp = res; res = invalid; return tmp; }
    bool valid() { return res != invalid; }

private:
    AutoRes(const AutoRes&);
    AutoRes& operator= (const AutoRes&);

private:
    T res;
    FreeRes free_res;
};

static inline int test_bit(const void* addr, int bit)
{
    return !!(((uint32_t*)addr)[bit >> 5] & (1 << (bit & 0x1f)));
}

static inline int test_bit_be(const void* addr, int bit)
{
    return !!(((uint8_t*)addr)[bit >> 3] & (0x80 >> (bit & 0x07)));
}

static inline void set_bit(const void* addr, int bit)
{
    ((uint8_t*)addr)[bit >> 5] |= (0x1 << (bit & 0x1f));
}

static inline void set_bit_be(const void* addr, int bit)
{
    ((uint8_t*)addr)[bit >> 3] |= (0x80 >> (bit & 0x07));
}

#ifdef WIN32

class Delete_DC {
public:
    void operator() (HDC dc) { DeleteDC(dc);}
};
typedef AutoRes<HDC, Delete_DC> AutoDC;

class Delete_Object {
public:
    void operator() (HGDIOBJ obj) { DeleteObject(obj);}
};
typedef AutoRes<HGDIOBJ, Delete_Object> AutoGDIObject;

class DeleteOGLContext {
public:
    void operator() (HGLRC ctx) { wglDeleteContext(ctx);}
};
typedef AutoRes<HGLRC, DeleteOGLContext> AutoOGLCtx;

HDC create_compatible_dc();
HBITMAP get_bitmap_res(int id);
HBITMAP get_alpha_bitmap_res(int id);


class WindowDC {
public:
    WindowDC(HWND window) : _window (window), _dc (GetDC(window)) {}
    ~WindowDC() { ReleaseDC(_window, _dc);}
    HDC operator* () { return _dc;}

private:
    HWND _window;
    HDC _dc;
};
typedef AutoRes<HDC, Delete_DC> AutoReleaseDC;

#endif

template<class T, class FreeRes = FreeObject<T> >
class _AutoRes {
public:
    _AutoRes() : _res (NULL) {}
    _AutoRes(T* res) : _res (res) {}
    ~_AutoRes() { set(NULL); }

    void set(T* res) {if (_res) _free_res(_res); _res = res; }
    T* get() {return _res;}
    T* operator -> () { return _res;}
    T* release() {T* tmp = _res; _res = NULL; return tmp; }
    bool valid() { return !!_res; }

private:
    _AutoRes(const _AutoRes&);
    _AutoRes& operator= (const _AutoRes&);

private:
    T* _res;
    FreeRes _free_res;
};

std::wstring char2wstring (const char *src);
#endif
