#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <errno.h>

#include "utils.h"
#include "xplatform.h"
#include "debug.h"
#include "rederrorcodes.h"

#include "controller.h"

const char *channel_names[] = { "dummy", "main", "display", "inputs", "cursor", "playback", "record" };

void QErrorHandler(int err, const char *custom_string)
{
    LOG_DEBUG("Something went wrong: " << custom_string << ", " << err);
}

SpiceController::SpiceController(SpiceConfiguration *config) 
{
    LOG_TRACE("");
    _spice_config = config;

    m_data = new MessageData();
    memset (m_data, 0, sizeof (m_data));
    char *p_socket = getenv("SPICE_XPI_SOCKET");
    if (p_socket) {
        strncpy (m_data->name, p_socket, sizeof(m_data->name));
    }

   _is_ipc_connect_recieved = false;
}

SpiceController::SpiceController(MessageData &data)
{
    LOG_TRACE("");

    _spice_config = NULL;
    m_data = new MessageData();
    memcpy (m_data, &data, sizeof (MessageData));
}

SpiceController::~SpiceController()
{
    LOG_TRACE("");

    DisconnectClient();
    Disconnect();
    free (m_data);
}

void SpiceController::Create()
{
    LOG_TRACE("");
    struct sockaddr_un local;

    if ((m_data->listen_socket = socket (AF_UNIX, SOCK_STREAM, 0)) == -1) {
        QErrorHandler(errno, "create socket error");
    }

    local.sun_family = AF_UNIX;
    strcpy(local.sun_path, m_data->name);
    unlink(local.sun_path);
    DBG(0, fmt("create %s") % m_data->name);
    if (bind(m_data->listen_socket, (struct sockaddr *)&local, strlen(local.sun_path) + sizeof (local.sun_family)) == -1) {
        QErrorHandler(errno, "bind error");
    }

    if (listen(m_data->listen_socket, 2) == -1) {
        QErrorHandler(errno, "listen error");
    }
}

int SpiceController::Connect()
{
    int nRC;

    LOG_TRACE("");
    struct sockaddr_un remote;

    if ((m_data->client_socket = socket (AF_UNIX, SOCK_STREAM, 0)) == -1) {
        QErrorHandler(errno, "SpiceController::Connect socket create error");
    }

    remote.sun_family = AF_UNIX;
    strcpy(remote.sun_path, m_data->name);

    DBG(0, fmt("connect to %s") % m_data->name);
    nRC = connect(m_data->client_socket, (struct sockaddr *)&remote, strlen(remote.sun_path) + sizeof (remote.sun_family));
    if (nRC == -1) {
        QErrorHandler(errno, "connect error");
        LOG_DEBUG("Connect Error");
    } else {
        LOG_DEBUG("Connected!");
    }

    return (nRC);
}

int SpiceController::Connect(const int nRetries)
{
    int nRC;
    int nCount = 0;

    do {
        sleep(1);
        nRC = Connect();
        nCount++;
    } while (nRC != 0 && nCount < nRetries);

    return (nRC);
}

void SpiceController::DisconnectClient()
{
    LOG_TRACE("");
    close (m_data->client_socket);
}

void SpiceController::Disconnect()
{
    LOG_TRACE("");
    close (m_data->listen_socket);
    unlink(m_data->name);
}

void SpiceController::Write(const void *lpBuffer, uint32_t nBytesToWrite, uint32_t *nBytesWritten)
{
    LOG_TRACE("");
    int len;

    len = send (
        m_data->client_socket,
        lpBuffer,
        nBytesToWrite,
        0
    );

    if (len != nBytesToWrite) {
    LOG_WARN("send error, bytes to write = " << nBytesToWrite << ", bytes actually written = " << len << ", errno=" << errno);
        QErrorHandler(1, "send error");
    }

    *nBytesWritten = len;
}

void SpiceController::Read(void *lpBuffer, uint32_t nBytesToRead, uint32_t *nBytesRead)
{
    LOG_TRACE("");
    int len;

    len = recv (
        m_data->client_socket,
        lpBuffer,
        nBytesToRead,
        0
    );

    DBG(0, fmt("after recv socket=%d len=%d, errno=%d errstr=%s") % m_data->client_socket % len % errno % strerror(errno));

    if (len != nBytesToRead) {
        QErrorHandler(1, "recv error");
    }

    *nBytesRead = len;
}

bool SpiceController::WaitForPipe(uint32_t nTimeout)
{
    LOG_TRACE("");
    return (true);
}

bool SpiceController::new_client_connection()
{
    LOG_TRACE("");
    struct sockaddr_un remote;
    socklen_t len;

    len = sizeof (remote);
    m_data->client_socket = accept (m_data->listen_socket, (struct sockaddr *)&remote, &len);

    DBG(0, fmt("after accept socket=%d, errno=%d errstr=%s") % m_data->client_socket % errno % strerror(errno));

    return (true);
}

bool SpiceController::WaitForConnectMsg(uint32_t nTimeout)
{
    LOG_TRACE("");
    while (!_is_ipc_connect_recieved) {
        Platform::wait_events();
    }

    return true;
}

void SpiceController::SetApp(Application *app)
{
    LOG_TRACE("");
    _app = app;
}

SpiceConfiguration *SpiceController::GetConfig()
{
    LOG_TRACE("");
    return (_spice_config);
}

void SpiceController::ProcessMessage(ControlMessage& msg)
{
    LOG_TRACE("");
        // Handle msgs
    switch (msg.id) {
    case MSG_CONNECT:
        LOG_DEBUG ("MSG_CONNECT _spice_config=" << _spice_config);
        _is_ipc_connect_recieved = true;

        if (_app != NULL) {
            _app->config(*_spice_config);
            _app->connect();
        }
        break;
    case MSG_SHOW:
        DBG(0, "MSG_SHOW");
        if (_app != NULL) {
            _app->external_show();
        } else {
            LOG_ERROR ("MSG_SHOW was called prior to Application initalization");
        }
        break;
    case MSG_host:
        DBG(0, "MSG_host");
        _spice_config->_host = msg.szData;
        break;
    case MSG_port:
        DBG(0, "MSG_port");
        _spice_config->_port = atoi(msg.szData);
        break;
    case MSG_sport:
        DBG(0, "MSG_sport");
        _spice_config->_sport = atoi(msg.szData);
        break;
    case MSG_full_screen:
        DBG(0, "MSG_full_screen");
        _spice_config->_full_screen = msg.szData[0];
        break;
    case MSG_no_audio:
        DBG(0, "MSG_no_audio");
        _spice_config->_no_audio = msg.szData[0];
    break;
    case MSG_auto_display_res:
        DBG(0, "MSG_auto_display_res");
        _spice_config->_auto_display_res = !msg.szData[0];
        break;
    case MSG_password:
        DBG(0, "MSG_password");
        _spice_config->_password = msg.szData;
        break;
    case MSG_cipher:
        DBG(0, "MSG_cipher");
        _spice_config->_cipher = msg.szData;
    break;
    case MSG_title:
        DBG(0, "MSG_title");
        _spice_config->_title = char2wstring(msg.szData);
        break;
    case MSG_encrypt_channels: {
        std::string encrypt_channels;
        DBG(0, "MSG_encrypt_channels");
        encrypt_channels = msg.szData;

        if (encrypt_channels.compare("all") == 0) {
            memset(_spice_config->_encrypt_channels, true, sizeof(_spice_config->_encrypt_channels));
        } else {
            for (int i = 0; i < RED_CHANNEL_END; ++i) {
                _spice_config->_encrypt_channels[i] = (encrypt_channels.find(channel_names[i]) != std::string::npos);
                if (encrypt_channels.find(channel_names[i]) != std::string::npos) {
                    DBG(0, fmt ("found %s") % channel_names[i]);
                }
            }
        }
        break;
    } 
    case MSG_hotkeys:
        DBG(0, "MSG_hotkeys");
        _spice_config->parse_hotkeys(msg.szData);
        break;
    case MSG_send_cad:
        DBG(0, "MSG_send_cad");
        _spice_config->_send_cad = msg.szData[0];
        break;
#ifdef WIN32
    case MSG_no_taskmgr,
        DBG(0, "MSG_no_taskmgr");
        _spice_config->_no_taskmgr = msg.szData[0];
        break;
#endif
    case MSG_CA_file:
        DBG(0, "MSG_CA_file");
        _spice_config->_CA_file = msg.szData;
    break;
    case MSG_host_cert_subject:
    {
        std::string subject = msg.szData;
        DBG(0, "MSG_host_cert_subject");   
        _spice_config->parse_host_cert_subject(subject);
        break;
    }
    default:
        DBG(0, fmt("unsuported %||") % msg.id);
    }
}

int SpiceController::TranslateRC (int nRC) {
    LOG_TRACE("");
    int nReturn = 0;

    switch (nRC) {
    case SPICEC_ERROR_CODE_SUCCESS:
        nReturn = 0;
        break;
    case SPICEC_ERROR_CODE_GETHOSTBYNAME_FAILED:
		nReturn = RDP_ERROR_CODE_HOST_NOT_FOUND;
		break;
    case SPICEC_ERROR_CODE_CONNECT_FAILED:
		nReturn = RDP_ERROR_CODE_WINSOCK_CONNECT_FAILED;
		break;
    case SPICEC_ERROR_CODE_ERROR:
    case SPICEC_ERROR_CODE_SOCKET_FAILED:
		nReturn = RDP_ERROR_CODE_INTERNAL_ERROR;
		break;
    case SPICEC_ERROR_CODE_RECV_FAILED:
        nReturn = RDP_ERROR_RECV_WINSOCK_FAILED;
        break;
    case SPICEC_ERROR_CODE_SEND_FAILED:
        nReturn = RDP_ERROR_SEND_WINSOCK_FAILED;
        break;
    case SPICEC_ERROR_CODE_NOT_ENOUGH_MEMORY:
        nReturn = RDP_ERROR_CODE_OUT_OF_MEMORY;
        break;
    case SPICEC_ERROR_CODE_AGENT_TIMEOUT:
        nReturn = RDP_ERROR_CODE_TIMEOUT;
        break;
    case SPICEC_ERROR_CODE_AGENT_ERROR:
        nReturn = RDP_ERROR_CODE_INTERNAL_ERROR;
        break;
    default:
        nReturn = RDP_ERROR_CODE_INTERNAL_ERROR;
    }

    return nReturn;
}

